from typing import Any

from i_dot_ai_utilities.logging.types.lambda_enrichment_schema import (
    ExtractedLambdaContext,
    LambdaContextLike,
)


class LambdaContextEnricher:
    def extract_context(
        self, logger: Any, request: LambdaContextLike
    ) -> ExtractedLambdaContext | None:
        response: ExtractedLambdaContext | None = None
        try:
            self._validate_object_instance(request)

            response = {
                "lambda_context": {
                    "request_id": request.aws_request_id,
                    "function_arn": request.invoked_function_arn,
                }
            }
        except Exception:
            logger.exception(
                "Exception(Logger): Failed to extract Lambda Context fields"
            )
            return None
        else:
            return response

    def _validate_object_instance(self, request: LambdaContextLike) -> None:
        if not isinstance(request, LambdaContextLike):
            msg = (
                "Exception(Logger): Object doesn't conform to "
                "LambdaContextLike. Context not set."
            )
            raise TypeError(msg)



from typing import TypedDict

from i_dot_ai_utilities.logging.enrichers.enrichment_provider import (
    ContextEnrichmentType,
)
from i_dot_ai_utilities.logging.enrichers.fastapi_enricher import RequestLike


class ContextEnrichmentOptions(TypedDict):
    type: ContextEnrichmentType
    object: RequestLike



import json
import sys
import time

from i_dot_ai_utilities.logging.structured_logger import StructuredLogger
from i_dot_ai_utilities.metrics.interfaces import MetricsWriter
from i_dot_ai_utilities.metrics.types.embedded_metric_format import (
    EmbeddedMetricFormat,
    StorageResolution,
)


class CloudwatchEmbeddedMetricsWriter(MetricsWriter):
    """Create a new CloudWatch Metrics Writer.

    Metrics are logged to stdout in the Embedded Metrics Format, which are automatically registered as time-series metrics by CloudWatch Logs.

    :param namespace: The namespace in CloudWatch in which to store all metrics. Usually the service/repo name, or some other app identifier.
    :param environment: The environment in which the code is running (e.g. dev/preprod/prod).
    :param logger: A Structured Logger used to emit messages on write failures.
    """  # noqa: E501

    def __init__(self, namespace: str, environment: str, logger: StructuredLogger):
        self.namespace = namespace
        self.environment = environment
        self._logger = logger

    def put_metric(
        self,
        metric_name: str,
        value: float,
        dimensions: dict | None = None,
    ) -> None:
        """Put a time-series metric to CloudWatch.

        See the i.AI utils readme for full details on usage.

        :param metric_name: The name of the metric to log.
        :param value: The numerical metric value.
        :param dimensions: A k/v set of **low-cardinality** dimensions to add to the metric for graphing purposes.
        """  # noqa: E501
        try:
            self._put_metric_internal(metric_name, value, dimensions)
        except Exception:
            self._logger.exception("Failed to write metric")

    def _put_metric_internal(
        self, metric_name: str, value: float, dimensions: dict | None = None
    ) -> None:
        if not metric_name or not value:
            msg = "Missing required parameter"
            raise ValueError(msg)

        if type(metric_name) is not str or type(value) not in [int, float]:
            msg = "Incorrect parameter type"
            raise ValueError(msg)

        dimensions = dimensions or {}
        dimension_names = list(dimensions.keys()) if dimensions else []

        emf: EmbeddedMetricFormat = {
            "_aws": {
                "Timestamp": int(time.time() * 1000),
                "CloudWatchMetrics": [
                    {
                        "Namespace": self.namespace + "/" + self.environment,
                        "Dimensions": [dimension_names] if dimension_names else [],
                        "Metrics": [
                            {
                                "Name": metric_name,
                                "Unit": "Count",
                                "StorageResolution": StorageResolution.STANDARD.value,
                            }
                        ],
                    }
                ],
            },
            **dimensions,
        }

        metric_payload = {**emf, metric_name: value}

        print(json.dumps(metric_payload), file=sys.stdout)  # noqa: T201



from abc import ABC, abstractmethod
from typing import Any

from i_dot_ai_utilities.logging.types.fargate_enrichment_schema import (
    ExtractedFargateContext,
)
from i_dot_ai_utilities.logging.types.lambda_enrichment_schema import (
    ExtractedLambdaEnvironmentMetadata,
)

ExtractedContextResult = (
    ExtractedFargateContext | ExtractedLambdaEnvironmentMetadata | None
)


class BaseEnvironmentEnricher(ABC):
    @abstractmethod
    def extract_context(self, self_logger: Any) -> ExtractedContextResult:
        pass



from typing import Protocol, TypedDict, runtime_checkable


@runtime_checkable
class LambdaContextLike(Protocol):
    @property
    def aws_request_id(self) -> str: ...

    @property
    def invoked_function_arn(self) -> str: ...


class LambdaContextMetadata(TypedDict):
    request_id: str
    function_arn: str


class ExtractedLambdaContext(TypedDict):
    lambda_context: LambdaContextMetadata


class LambdaEnvironmentMetadata(TypedDict):
    aws_region: str
    function_name: str


class ExtractedLambdaEnvironmentMetadata(TypedDict):
    lambda_os: LambdaEnvironmentMetadata



import logging
import os
import uuid
from typing import TYPE_CHECKING, Any

import structlog

from i_dot_ai_utilities.logging.enrichers.context_extractor import (
    ExtractedContextResult,
)
from i_dot_ai_utilities.logging.enrichers.enrichment_provider import (
    EnrichmentProvider,
    ExecutionEnvironmentType,
)
from i_dot_ai_utilities.logging.processor_helper import ProcessorHelper
from i_dot_ai_utilities.logging.types.context_enrichment_options import (
    ContextEnrichmentOptions,
)
from i_dot_ai_utilities.logging.types.context_fields import ContextFieldValue
from i_dot_ai_utilities.logging.types.log_output_format import LogOutputFormat
from i_dot_ai_utilities.logging.types.logger_config_options import LoggerConfigOptions

if TYPE_CHECKING:
    from i_dot_ai_utilities.logging.types.base_context import BaseContext
    from i_dot_ai_utilities.logging.types.fastapi_enrichment_schema import (
        ExtractedFastApiContext,
    )


class StructuredLogger:
    """Create a new Structured Logger.

    Logs can be output in either JSON or Console format. The execution environment can also be set to enrich log messages with environment-specific context.

    See the i.AI utils readme for full details on usage.

    :param level: The logging level. Log messages raised at this urgency or above will be written to stdout. Defaults to 'INFO'. Example values: 'DEBUG' | 'INFO' | 'WARNING' | 'ERROR'
    :param options: A set of LoggerConfigOptions that can be used to modify the logger behaviour.
    """  # noqa: E501

    _logger: Any
    _default_config: LoggerConfigOptions
    _execution_environment: ExecutionEnvironmentType
    _enricher_provider: EnrichmentProvider
    _log_format: LogOutputFormat
    _ship_logs: bool
    _logger_name: str | None = None

    def __init__(
        self,
        level: str | int = logging.INFO,
        options: LoggerConfigOptions | None = None,
    ):
        self._default_config = self._load_config_defaults()
        if not options:
            options = self._default_config

        self._logger = structlog.get_logger()

        self._logger_name = options.get("logger_name", None)

        self._execution_environment = options.get(
            "execution_environment", self._default_config["execution_environment"]
        )
        self._enricher_provider = EnrichmentProvider(self._execution_environment)

        self._log_format = options.get("log_format", self._default_config["log_format"])
        self._ship_logs = self._should_ship_logs(options)

        ProcessorHelper().configure_processors(
            self._normalise_log_level(level),
            self._log_format,
        )

        self._upsert_base_context()

    def debug(self, message_template: str, **kwargs: Any) -> None:
        """Write a debug log message.

        Log messages may be string-literals, or be formatted strings. Formatted strings will have any interpolated values added as context fields to the log message.

        Usage Examples
        ----------
        logger.debug('User started login flow', id=12345) # Log output: {"message": "User started login flow", "id": 12345}

        logger.debug('User {email} has logged in', email='me@example.com') # Log output: {"message": "User me@example.com has logged in", "email": "me@example.com"}

        :param message_template: The string literal or formatted string to pass to the logger.
        :param **kwargs: Arguments passed to interpolate into a formatted string, if using.
        """  # noqa: E501
        message = self._get_interpolated_message(message_template, **kwargs)
        self._logger.debug(message, message_template=message_template, **kwargs)

    def info(self, message_template: str, **kwargs: Any) -> None:
        """Write an informational log message.

        Log messages may be string-literals, or be formatted strings. Formatted strings will have any interpolated values added as context fields to the log message.

        Usage Examples
        ----------
        logger.info('User started login flow', id=12345) # Log output: {"message": "User started login flow", "id": 12345}

        logger.info('User {email} has logged in', email='me@example.com') # Log output: {"message": "User me@example.com has logged in", "email": "me@example.com"}

        :param message_template: The string literal or formatted string to pass to the logger.
        :param **kwargs: Arguments passed to interpolate into a formatted string, if using.
        """  # noqa: E501
        message = self._get_interpolated_message(message_template, **kwargs)
        self._logger.info(message, message_template=message_template, **kwargs)

    def warning(self, message_template: str, **kwargs: Any) -> None:
        """Write a warning log message.

        Log messages may be string-literals, or be formatted strings. Formatted strings will have any interpolated values added as context fields to the log message.

        Usage Examples
        ----------
        logger.warning('anonymous user failed to do X action', id=12345)
        => Log output: {"message": "anonymous user failed to do X action", "id": 12345}

        logger.warning('User {email} failed to log in due to password mismatch', email='me@example.com')
        => Log output: {"message": "User me@example.com failed to log in due to password mismatch", "email": "me@example.com"}

        :param message_template: The string literal or formatted string to pass to the logger.
        :param **kwargs: Arguments passed to interpolate into a formatted string, if using.
        """  # noqa: E501
        message = self._get_interpolated_message(message_template, **kwargs)
        self._logger.warning(message, message_template=message_template, **kwargs)

    def error(self, message_template: str, **kwargs: Any) -> None:
        """Write an error log message.

        Log messages may be string-literals, or be formatted strings. Formatted strings will have any interpolated values added as context fields to the log message.

        Usage Examples
        ----------
        logger.error('An error occurred with user login flow', id=12345) # Log output: {"message": "User started login flow", "id": 12345}

        logger.error('Logout failed for user {email}', email='me@example.com') # Log output: {"message": "Logout failed for user me@example.com", "email": "me@example.com"}

        :param message_template: The string literal or formatted string to pass to the logger.
        :param **kwargs: Arguments passed to interpolate into a formatted string, if using.
        """  # noqa: E501
        message = self._get_interpolated_message(message_template, **kwargs)
        self._logger.error(message, message_template=message_template, **kwargs)

    def exception(self, message_template: str, **kwargs: Any) -> None:
        """Write a caught exception, along with an error log message. Caught exceptions will automatically be added as context to the log message.

        Log messages may be string-literals, or be formatted strings. Formatted strings will have any interpolated values added as context fields to the log message.

        Usage Examples
        ----------
        logger.exception('Login flow failed for user', id=12345) # Log output: {"message": "Login flow failed for user", "id": 12345, "exception": "Traceback ..."}

        logger.exception('User {email} failed to update in DB', email='me@example.com') # Log output: {"message": User me@example.com failed to update in DB", "email": "me@example.com", "exception": "Traceback ..."}
        """  # noqa: E501
        message = self._get_interpolated_message(message_template, **kwargs)
        self._logger.exception(message, message_template=message_template, **kwargs)

    def set_context_field(self, field_key: str, field_value: ContextFieldValue) -> None:
        """Add a custom field to the logger dictionary. This field will appear on subsequent log messages.

        This key and value will be made available as a searchable field in the downstream logging stack.

        A field set using this function will be removed upon a refresh of the logger context.

        :param field_key: The key of the field.
        :param field_value: The value of the field.
        """  # noqa: E501
        structlog.contextvars.bind_contextvars(**{field_key: field_value})

    def refresh_context(
        self, context_enrichers: list[ContextEnrichmentOptions] | None = None
    ) -> None:
        """Reset the logger, creating a new context id and removing any custom fields set since the previous invocation.

        :param context_enrichers: A list of one or more ContextEnrichmentOptions. Used to refresh the new logger with fields from well-known frameworks, such as FastAPI request metadata.
        """  # noqa: E501
        structlog.contextvars.clear_contextvars()
        self._upsert_base_context()

        if context_enrichers is None:
            return

        additional_context: ExtractedFastApiContext | dict[str, Any] = {}
        for enricher in context_enrichers:
            enricher_type = enricher["type"]
            enricher_object = enricher["object"]
            ctx = self._enricher_provider.extract_context_from_framework_enricher(
                self, enricher_type, enricher_object
            )

            if ctx is not None:
                additional_context.update(ctx)

        structlog.contextvars.bind_contextvars(**additional_context)

    def _should_ship_logs(self, options: LoggerConfigOptions) -> bool:
        selected_option = options.get("ship_logs", self._default_config["ship_logs"])
        if (
            options.get("log_format", self._default_config["log_format"])
            is not LogOutputFormat.JSON
            and selected_option is True
        ):
            self._logger.warning(
                "Warning(Logger): messages cannot be shipped downstream "
                "outside of JSON format. Disabling log shipping"
            )
            return False

        return selected_option

    def _get_interpolated_message(self, message_template: str, **kwargs: Any) -> str:
        try:
            return message_template.format(**kwargs)
        except KeyError:
            self._logger.exception(
                (
                    "Exception(Logger): Variable interpolation failed when formatting "
                    "log message. Is a value missing?"
                ),
                message_template=message_template,
            )
            return message_template

    def _set_environment_context(
        self, environment_context: ExtractedContextResult
    ) -> None:
        if environment_context:
            structlog.contextvars.bind_contextvars(**environment_context)

    def _set_logger_name(self) -> None:
        if self._logger_name:
            structlog.contextvars.bind_contextvars(
                logger_name=self._logger_name,
            )

    def _upsert_base_context(self) -> None:
        self._set_environment_context(
            self._enricher_provider.load_execution_environment_context(self)
        )

        self._set_logger_name()

        base_context: BaseContext = {
            "context_id": str(uuid.uuid4()),
            "env": {
                "app_name": os.environ.get("APP_NAME", "unknown"),
                "repo_name": os.environ.get("REPO", "unknown"),
                "environment_name": os.environ.get("ENVIRONMENT", "unknown"),
            },
            "ship_logs": 1 if self._ship_logs else 0,
        }
        structlog.contextvars.bind_contextvars(**base_context)

    def _normalise_log_level(self, level: str | int) -> int:
        match level.upper() if isinstance(level, str) else level:
            case "DEBUG" | logging.DEBUG:
                return logging.DEBUG
            case "INFO" | "INFORMATION" | "INFORMATIONAL" | logging.INFO:
                return logging.INFO
            case "WARN" | "WARNING" | logging.WARNING:
                return logging.WARNING
            case "ERROR" | logging.ERROR:
                return logging.ERROR
            case _:
                self._logger.warning(
                    "Log level {level} not recognised, defaulting to INFO", level=level
                )
                return logging.INFO

    def _load_config_defaults(self) -> LoggerConfigOptions:
        return {
            "execution_environment": ExecutionEnvironmentType.FARGATE,
            "log_format": LogOutputFormat.JSON,
            "ship_logs": True,
        }



from enum import Enum
from typing import TypedDict


class StorageResolution(Enum):
    """Sets the resolution at which metrics are stored in CloudWatch.

    :Standard: Graphing available at 60-second granularity (per-minute datapoints)
    :High: Stored at 1-second granularity (per-second datapoints)
    """

    STANDARD = 60
    HIGH = 1


class MetricDefinition(TypedDict):
    Name: str
    Unit: str
    StorageResolution: StorageResolution | int


class CloudWatchMetricBlock(TypedDict):
    Namespace: str
    Dimensions: list[list[str]]
    Metrics: list[MetricDefinition]


class AWSBlock(TypedDict):
    Timestamp: int
    CloudWatchMetrics: list[CloudWatchMetricBlock]


class EmbeddedMetricFormat(TypedDict):
    _aws: AWSBlock



from functools import lru_cache
from typing import Any

from pydantic import Field
from pydantic_settings import BaseSettings

from i_dot_ai_utilities.logging.enrichers.context_extractor import (
    BaseEnvironmentEnricher,
)
from i_dot_ai_utilities.logging.types.lambda_enrichment_schema import (
    ExtractedLambdaEnvironmentMetadata,
)


class LambdaEnvironmentSettings(BaseSettings):
    aws_default_region: str = Field()
    aws_lambda_function_name: str = Field()


@lru_cache
def load_lambda_environment_variables() -> LambdaEnvironmentSettings:
    return LambdaEnvironmentSettings()  # type: ignore[call-arg]


class LambdaEnvironmentEnricher(BaseEnvironmentEnricher):
    def extract_context(
        self, self_logger: Any
    ) -> ExtractedLambdaEnvironmentMetadata | None:
        response: ExtractedLambdaEnvironmentMetadata | None = None
        try:
            loaded_metadata = load_lambda_environment_variables()

            response = {
                "lambda_os": {
                    "aws_region": loaded_metadata.aws_default_region,
                    "function_name": loaded_metadata.aws_lambda_function_name,
                }
            }
        except Exception:
            self_logger.exception(
                "Exception(Logger): Failed to extract Lambda environment variables"
            )
            return None
        else:
            return response



from abc import ABC, abstractmethod


class MetricsWriter(ABC):
    @abstractmethod
    def put_metric(
        self,
        metric_name: str,
        value: float,
        dimensions: dict | None = None,
    ) -> None:
        pass



from typing import TypedDict


class FargateContainerLabelsLike:
    def __init__(self, raw_labels: dict[str, str]):
        self._raw = raw_labels

    def __getitem__(self, key: str) -> str:
        return self._raw[key]

    @property
    def task_arn(self) -> str:
        return self._raw.get("com.amazonaws.ecs.task-arn", "unknown")


class FargateMetadata(TypedDict):
    image_id: str
    task_arn: str
    container_started_at: str
    aws_region: str


class ExtractedFargateContext(TypedDict):
    fargate: FargateMetadata



from typing import Any, Protocol, TypedDict, runtime_checkable


@runtime_checkable
class URLLike(Protocol):
    @property
    def path(self) -> str: ...

    @property
    def query(self) -> str: ...


@runtime_checkable
class RequestLike(Protocol):
    @property
    def method(self) -> str: ...

    @property
    def path_params(self) -> dict[str, Any]: ...

    @property
    def base_url(self) -> Any: ...

    @property
    def headers(self) -> Any: ...

    @property
    def url(self) -> URLLike: ...


class FastApiRequestMetadata(TypedDict):
    method: str
    base_url: str
    user_agent: str
    x_forwarded_for: str
    path: str
    query: str


class ExtractedFastApiContext(TypedDict):
    request: FastApiRequestMetadata



from typing import NotRequired, TypedDict

from i_dot_ai_utilities.logging.enrichers.enrichment_provider import (
    ExecutionEnvironmentType,
)
from i_dot_ai_utilities.logging.types.log_output_format import LogOutputFormat


class LoggerConfigOptions(TypedDict):
    execution_environment: NotRequired[ExecutionEnvironmentType]
    log_format: NotRequired[LogOutputFormat]
    ship_logs: NotRequired[bool]
    logger_name: NotRequired[str]



from collections.abc import Callable

import structlog
from structlog.dev import ConsoleRenderer
from structlog.processors import JSONRenderer

from i_dot_ai_utilities.logging.types.log_output_format import LogOutputFormat


class ProcessorHelper:
    def configure_processors(self, log_level: int, log_format: LogOutputFormat) -> None:
        structlog.configure(
            wrapper_class=structlog.make_filtering_bound_logger(log_level)
        )
        structlog.configure(
            processors=[
                structlog.contextvars.merge_contextvars,
                structlog.processors.TimeStamper(fmt="iso"),
                structlog.processors.add_log_level,
                structlog.processors.EventRenamer("message"),
                *self._load_output_processors(log_format),
            ]
        )

    def _load_output_processors(
        self, output_type: LogOutputFormat
    ) -> tuple[Callable, ...]:
        match output_type:
            case LogOutputFormat.JSON:
                return (
                    structlog.processors.format_exc_info,
                    JSONRenderer(),
                )
            case _:
                return (ConsoleRenderer(),)



from collections.abc import Sequence

ContextFieldPrimitives = str | int | bool | float
ContextFieldValue = (
    ContextFieldPrimitives
    | Sequence[ContextFieldPrimitives]
    | dict[str, ContextFieldPrimitives]
)



import json
import os
from typing import Any
from urllib import request
from urllib.parse import urlparse

from i_dot_ai_utilities.logging.enrichers.context_extractor import (
    BaseEnvironmentEnricher,
)
from i_dot_ai_utilities.logging.types.fargate_enrichment_schema import (
    ExtractedFargateContext,
    FargateContainerLabelsLike,
)


class FargateEnvironmentEnricher(BaseEnvironmentEnricher):
    _container_metadata_url_parameter_name: str = "ECS_CONTAINER_METADATA_URI_V4"

    def extract_context(self, self_logger: Any) -> ExtractedFargateContext | None:
        response: ExtractedFargateContext | None = None
        try:
            metadata_response = self._get_metadata_response()
            loaded_metadata = FargateContainerMetadataResponse(metadata_response)

            response = {
                "fargate": {
                    "image_id": loaded_metadata.image_id,
                    "task_arn": loaded_metadata.labels.task_arn,
                    "container_started_at": loaded_metadata.started_at,
                    "aws_region": loaded_metadata.aws_region,
                }
            }
        except Exception:
            self_logger.exception(
                "Exception(Logger): Failed to extract Fargate container metadata fields"
            )
            return None
        else:
            return response

    def _get_metadata_response(self) -> Any:
        url = os.environ.get(self._container_metadata_url_parameter_name, None)

        if url is None:
            msg = "Failed to find metadata URL on environment"
            raise ValueError(msg)

        parsed_url = urlparse(url)
        if parsed_url.scheme not in ("http", "https"):
            msg = "URL must use HTTP or HTTPS"
            raise ValueError(msg)

        with request.urlopen(parsed_url.geturl()) as response:  # noqa: S310 # nosec B310
            return json.loads(response.read())


class FargateContainerMetadataResponse:
    def __init__(self, raw_response: dict[str, Any]):
        try:
            self.image_id: str = raw_response["ImageID"]
            self.started_at: str = raw_response["StartedAt"]
            self.labels: FargateContainerLabelsLike = FargateContainerLabelsLike(
                raw_response["Labels"]
            )
            self.aws_region: str = raw_response["AvailabilityZone"][:-1]

        except Exception as e:
            msg = (
                "Exception(Logger): Response doesn't conform to "
                "FargateContainerMetadataResponse. Context not set."
            )
            raise TypeError(msg) from e



from typing import Any

from i_dot_ai_utilities.logging.types.fastapi_enrichment_schema import (
    ExtractedFastApiContext,
    RequestLike,
)


class FastApiEnricher:
    def extract_context(
        self, logger: Any, request: RequestLike
    ) -> ExtractedFastApiContext | None:
        response: ExtractedFastApiContext | None = None
        try:
            self._validate_object_instance(request)

            response = {
                "request": {
                    "method": request.method,
                    "base_url": str(request.base_url),
                    "user_agent": request.headers.get("user-agent", "none"),
                    "x_forwarded_for": request.headers.get("x-forwarded-for", "none"),
                    "path": request.url.path,
                    "query": request.url.query,
                }
            }
        except Exception:
            logger.exception("Exception(Logger): Failed to extract FastAPI fields")
            return None
        else:
            return response

    def _validate_object_instance(self, request: RequestLike) -> None:
        if not isinstance(request, RequestLike):
            msg = (
                "Exception(Logger): Request object doesn't conform to "
                "RequestLike. Context not set."
            )
            raise TypeError(msg)



from typing import Any, cast

from i_dot_ai_utilities.logging.enrichers.context_extractor import (
    ExtractedContextResult,
)
from i_dot_ai_utilities.logging.enrichers.fargate_enricher import (
    FargateEnvironmentEnricher,
)
from i_dot_ai_utilities.logging.enrichers.fastapi_enricher import (
    FastApiEnricher,
    RequestLike,
)
from i_dot_ai_utilities.logging.enrichers.lambda_context_enricher import (
    LambdaContextEnricher,
)
from i_dot_ai_utilities.logging.enrichers.lambda_environment_enricher import (
    LambdaEnvironmentEnricher,
)
from i_dot_ai_utilities.logging.types.enrichment_types import (
    ContextEnrichmentType,
    ExecutionEnvironmentType,
)
from i_dot_ai_utilities.logging.types.fastapi_enrichment_schema import (
    ExtractedFastApiContext,
)
from i_dot_ai_utilities.logging.types.lambda_enrichment_schema import (
    ExtractedLambdaContext,
    LambdaContextLike,
)


class EnrichmentProvider:
    _fast_api_enricher: FastApiEnricher
    _lambda_enricher: LambdaContextEnricher
    _execution_environment_enricher: (
        FargateEnvironmentEnricher | LambdaEnvironmentEnricher | None
    )
    _execution_environment_context_cache: ExtractedContextResult = None
    _has_environment_context_extraction_ran = False

    def __init__(self, execution_environment: ExecutionEnvironmentType):
        self._fast_api_enricher = FastApiEnricher()
        self._lambda_enricher = LambdaContextEnricher()

        match execution_environment:
            case ExecutionEnvironmentType.FARGATE:
                self._execution_environment_enricher = FargateEnvironmentEnricher()
            case ExecutionEnvironmentType.LAMBDA:
                self._execution_environment_enricher = LambdaEnvironmentEnricher()
            case _:
                self._execution_environment_enricher = None

    def extract_context_from_framework_enricher(
        self,
        self_logger: Any,
        enricher_type: ContextEnrichmentType,
        enricher_object: RequestLike | LambdaContextLike,
    ) -> ExtractedFastApiContext | ExtractedLambdaContext | None:
        match enricher_type:
            case ContextEnrichmentType.FASTAPI:
                return self._fast_api_enricher.extract_context(
                    self_logger, cast("RequestLike", enricher_object)
                )
            case ContextEnrichmentType.LAMBDA:
                return self._lambda_enricher.extract_context(
                    self_logger, cast("LambdaContextLike", enricher_object)
                )
            case _:
                self_logger.exception(
                    (
                        "Exception(Logger): An enricher type of '{enricher_type}' was "
                        "not recognised, no context added."
                    ),
                    enricher_type=enricher_type,
                )
                return None

    def load_execution_environment_context(
        self, self_logger: Any
    ) -> ExtractedContextResult:
        if self._execution_environment_enricher is None:
            return None

        if self._has_environment_context_extraction_ran:
            return self._execution_environment_context_cache

        self._execution_environment_context_cache = (
            self._execution_environment_enricher.extract_context(self_logger)
        )
        self._has_environment_context_extraction_ran = True
        return self._execution_environment_context_cache



from typing import Literal, TypedDict


class BaseContextEnvVars(TypedDict):
    app_name: str
    repo_name: str
    environment_name: str


class BaseContext(TypedDict):
    context_id: str
    env: BaseContextEnvVars
    ship_logs: Literal[1, 0]
